import { Role } from './ad.types';
import { Target } from './candidate.types';
import { CarpoolPathCreatorException } from './match.errors';
import { Actor } from './value-objects/actor.value-object';
import { Point } from './value-objects/point.value-object';
import { WayStep } from './value-objects/waystep.value-object';

export class CarpoolPathCreator {
  private PRECISION = 5;

  constructor(
    private readonly driverWaypoints: Point[],
    private readonly passengerWaypoints: Point[],
  ) {
    if (driverWaypoints.length < 2)
      throw new CarpoolPathCreatorException(
        new Error('At least 2 driver waypoints must be defined'),
      );
    if (passengerWaypoints.length < 2)
      throw new CarpoolPathCreatorException(
        new Error('At least 2 passenger waypoints must be defined'),
      );
  }

  /**
   * Creates a path (a list of waysteps) between driver waypoints
      and passenger waypoints respecting the order
      of the driver waypoints
      Inspired by :
      https://stackoverflow.com/questions/849211/shortest-distance-between-a-point-and-a-line-segment
   */
  public carpoolPath = (): WayStep[] =>
    this._consolidate(
      this._mixedWaysteps(this._driverWaysteps(), this._passengerWaysteps()),
    );

  private _mixedWaysteps = (
    driverWaysteps: WayStep[],
    passengerWaysteps: WayStep[],
  ): WayStep[] =>
    driverWaysteps.length == 2
      ? this._simpleMixedWaysteps(driverWaysteps, passengerWaysteps)
      : this._complexMixedWaysteps(driverWaysteps, passengerWaysteps);

  private _driverWaysteps = (): WayStep[] =>
    this.driverWaypoints.map(
      (waypoint: Point, index: number) =>
        new WayStep({
          point: new Point({
            lon: waypoint.lon,
            lat: waypoint.lat,
          }),
          actors: [
            new Actor({
              role: Role.DRIVER,
              target: this._getTarget(index, this.driverWaypoints),
            }),
          ],
        }),
    );

  /**
   * Creates the passenger waysteps with original passenger waypoints, adding driver waypoints that are the same
   */
  private _passengerWaysteps = (): WayStep[] => {
    const waysteps: WayStep[] = [];
    this.passengerWaypoints.forEach(
      (passengerWaypoint: Point, index: number) => {
        const waystep: WayStep = new WayStep({
          point: new Point({
            lon: passengerWaypoint.lon,
            lat: passengerWaypoint.lat,
          }),
          actors: [
            new Actor({
              role: Role.PASSENGER,
              target: this._getTarget(index, this.passengerWaypoints),
            }),
          ],
        });
        if (
          this.driverWaypoints.filter((driverWaypoint: Point) =>
            passengerWaypoint.isSame(driverWaypoint),
          ).length == 0
        ) {
          waystep.actors.push(
            new Actor({
              role: Role.DRIVER,
              target: Target.NEUTRAL,
            }),
          );
        }
        waysteps.push(waystep);
      },
    );
    return waysteps;
  };

  private _simpleMixedWaysteps = (
    driverWaysteps: WayStep[],
    passengerWaysteps: WayStep[],
  ): WayStep[] => [driverWaysteps[0], ...passengerWaysteps, driverWaysteps[1]];

  private _complexMixedWaysteps = (
    driverWaysteps: WayStep[],
    passengerWaysteps: WayStep[],
  ): WayStep[] => {
    let mixedWaysteps: WayStep[] = [...driverWaysteps];
    const originInsertIndex: number = this._insertIndex(
      passengerWaysteps[0],
      driverWaysteps,
    );
    mixedWaysteps = [
      ...mixedWaysteps.slice(0, originInsertIndex),
      passengerWaysteps[0],
      ...mixedWaysteps.slice(originInsertIndex),
    ];
    const destinationInsertIndex: number =
      this._insertIndex(
        passengerWaysteps[passengerWaysteps.length - 1],
        driverWaysteps,
      ) + 1;
    mixedWaysteps = [
      ...mixedWaysteps.slice(0, destinationInsertIndex),
      passengerWaysteps[passengerWaysteps.length - 1],
      ...mixedWaysteps.slice(destinationInsertIndex),
    ];
    return mixedWaysteps;
  };

  private _insertIndex = (
    targetWaystep: WayStep,
    waysteps: WayStep[],
  ): number =>
    this._closestSegmentIndex(targetWaystep, this._segments(waysteps)) + 1;

  private _segments = (waysteps: WayStep[]): WayStep[][] => {
    const segments: WayStep[][] = [];
    waysteps.forEach((waystep: WayStep, index: number) => {
      if (index < waysteps.length - 1)
        segments.push([waystep, waysteps[index + 1]]);
    });
    return segments;
  };

  private _closestSegmentIndex = (
    waystep: WayStep,
    segments: WayStep[][],
  ): number => {
    const distances: Map<number, number> = new Map();
    segments.forEach((segment: WayStep[], index: number) => {
      distances.set(index, this._distanceToSegment(waystep, segment));
    });
    const sortedDistances: Map<number, number> = new Map(
      [...distances.entries()].sort((a, b) => a[1] - b[1]),
    );
    const [closestSegmentIndex] = sortedDistances.keys();
    return closestSegmentIndex;
  };

  private _distanceToSegment = (waystep: WayStep, segment: WayStep[]): number =>
    parseFloat(
      Math.sqrt(this._distanceToSegmentSquared(waystep, segment)).toFixed(
        this.PRECISION,
      ),
    );

  private _distanceToSegmentSquared = (
    waystep: WayStep,
    segment: WayStep[],
  ): number => {
    const length2: number = this._distanceSquared(
      segment[0].point,
      segment[1].point,
    );
    if (length2 == 0)
      return this._distanceSquared(waystep.point, segment[0].point);
    const length: number = Math.max(
      0,
      Math.min(
        1,
        ((waystep.point.lon - segment[0].point.lon) *
          (segment[1].point.lon - segment[0].point.lon) +
          (waystep.point.lat - segment[0].point.lat) *
            (segment[1].point.lat - segment[0].point.lat)) /
          length2,
      ),
    );
    const newPoint: Point = new Point({
      lon:
        segment[0].point.lon +
        length * (segment[1].point.lon - segment[0].point.lon),
      lat:
        segment[0].point.lat +
        length * (segment[1].point.lat - segment[0].point.lat),
    });
    return this._distanceSquared(waystep.point, newPoint);
  };

  private _distanceSquared = (point1: Point, point2: Point): number =>
    parseFloat(
      (
        Math.pow(point1.lon - point2.lon, 2) +
        Math.pow(point1.lat - point2.lat, 2)
      ).toFixed(this.PRECISION),
    );

  private _getTarget = (index: number, waypoints: Point[]): Target =>
    index == 0
      ? Target.START
      : index == waypoints.length - 1
      ? Target.FINISH
      : Target.INTERMEDIATE;

  /**
   * Consolidate waysteps by removing duplicate actors (eg. driver with neutral and start or finish target)
   */
  private _consolidate = (waysteps: WayStep[]): WayStep[] => {
    const uniquePoints: Point[] = [];
    waysteps.forEach((waystep: WayStep) => {
      if (
        uniquePoints.find((point: Point) => point.isSame(waystep.point)) ===
        undefined
      )
        uniquePoints.push(
          new Point({
            lon: waystep.point.lon,
            lat: waystep.point.lat,
          }),
        );
    });
    return uniquePoints.map(
      (point: Point) =>
        new WayStep({
          point,
          actors: waysteps
            .filter((waystep: WayStep) => waystep.point.isSame(point))
            .map((waystep: WayStep) => waystep.actors)
            .flat(),
        }),
    );
  };
}
