import { Inject, Injectable } from '@nestjs/common';
import {
  DateTimeTransformerPort,
  Frequency,
  GeoDateTime,
} from '../core/application/ports/datetime-transformer.port';
import { TimeConverterPort } from '../core/application/ports/time-converter.port';
import {
  PARAMS_PROVIDER,
  TIMEZONE_FINDER,
  TIME_CONVERTER,
} from '../ad.di-tokens';
import { TimezoneFinderPort } from '../core/application/ports/timezone-finder.port';
import { DefaultParamsProviderPort } from '../core/application/ports/default-params-provider.port';

@Injectable()
export class InputDateTimeTransformer implements DateTimeTransformerPort {
  private readonly _defaultTimezone: string;
  constructor(
    @Inject(PARAMS_PROVIDER)
    private readonly defaultParamsProvider: DefaultParamsProviderPort,
    @Inject(TIMEZONE_FINDER)
    private readonly timezoneFinder: TimezoneFinderPort,
    @Inject(TIME_CONVERTER) private readonly timeConverter: TimeConverterPort,
  ) {
    this._defaultTimezone = defaultParamsProvider.getParams().TIMEZONE;
  }

  /**
   * Compute the fromDate : if an ad is punctual, the departure date
   * is converted to UTC with the time and timezone
   */
  fromDate = (geoFromDate: GeoDateTime, frequency: Frequency): string => {
    if (frequency === Frequency.RECURRENT) return geoFromDate.date;
    return this.timeConverter
      .localStringDateTimeToUtcDate(
        geoFromDate.date,
        geoFromDate.time,
        this.timezoneFinder.timezones(
          geoFromDate.coordinates.lon,
          geoFromDate.coordinates.lat,
          this._defaultTimezone,
        )[0],
      )
      .toISOString()
      .split('T')[0];
  };

  /**
   * Get the toDate depending on frequency, time and timezone :
   * if the ad is punctual, the toDate is equal to the fromDate
   */
  toDate = (
    toDate: string,
    geoFromDate: GeoDateTime,
    frequency: Frequency,
  ): string => {
    if (frequency === Frequency.RECURRENT) return toDate;
    return this.fromDate(geoFromDate, frequency);
  };

  /**
   * Get the day for a schedule item :
   * - if the ad is punctual, the day is infered from fromDate
   * - if the ad is recurrent, the day is computed by converting the time to utc
   */
  day = (
    day: number,
    geoFromDate: GeoDateTime,
    frequency: Frequency,
  ): number => {
    if (frequency === Frequency.RECURRENT)
      return this.recurrentDay(
        day,
        geoFromDate.time,
        this.timezoneFinder.timezones(
          geoFromDate.coordinates.lon,
          geoFromDate.coordinates.lat,
          this._defaultTimezone,
        )[0],
      );
    return new Date(this.fromDate(geoFromDate, frequency)).getDay();
  };

  /**
   * Get the utc time
   */
  time = (geoFromDate: GeoDateTime, frequency: Frequency): string => {
    if (frequency === Frequency.RECURRENT)
      return this.timeConverter.localStringTimeToUtcStringTime(
        geoFromDate.time,
        this.timezoneFinder.timezones(
          geoFromDate.coordinates.lon,
          geoFromDate.coordinates.lat,
          this._defaultTimezone,
        )[0],
      );
    return this.timeConverter
      .localStringDateTimeToUtcDate(
        geoFromDate.date,
        geoFromDate.time,
        this.timezoneFinder.timezones(
          geoFromDate.coordinates.lon,
          geoFromDate.coordinates.lat,
          this._defaultTimezone,
        )[0],
      )
      .toISOString()
      .split('T')[1]
      .split(':', 2)
      .join(':');
  };

  /**
   * Get the day for a schedule item for a recurrent ad
   * The day may change when transforming from local timezone to utc
   */
  private recurrentDay = (
    day: number,
    time: string,
    timezone: string,
  ): number => {
    const unixEpochDay = 4; // 1970-01-01 is a thursday !
    const utcBaseDay = this.timeConverter.utcUnixEpochDayFromTime(
      time,
      timezone,
    );
    if (unixEpochDay == utcBaseDay) return day;
    if (unixEpochDay > utcBaseDay) return day > 0 ? day - 1 : 6;
    return day < 6 ? day + 1 : 0;
  };
}
