import { Role } from '@modules/ad/core/domain/ad.types';
import { CandidateEntity } from '@modules/ad/core/domain/candidate.entity';
import { Target } from '@modules/ad/core/domain/candidate.types';
import { Actor } from '@modules/ad/core/domain/value-objects/actor.value-object';
import { Point } from '@modules/ad/core/domain/value-objects/point.value-object';

describe('Candidate entity', () => {
  it('should create a new candidate entity', () => {
    const candidateEntity: CandidateEntity = CandidateEntity.create({
      id: 'cc260669-1c6d-441f-80a5-19cd59afb777',
      role: Role.DRIVER,
      driverWaypoints: [
        {
          lat: 48.678454,
          lon: 6.189745,
        },
        {
          lat: 48.84877,
          lon: 2.398457,
        },
      ],
      passengerWaypoints: [
        {
          lat: 48.689445,
          lon: 6.17651,
        },
        {
          lat: 48.8566,
          lon: 2.3522,
        },
      ],
      driverDistance: 350145,
      driverDuration: 13548,
      driverSchedule: [
        {
          day: 0,
          time: '07:00',
          margin: 900,
        },
      ],
      passengerSchedule: [
        {
          day: 0,
          time: '07:10',
          margin: 900,
        },
      ],
      spacetimeDetourRatio: {
        maxDistanceDetourRatio: 0.3,
        maxDurationDetourRatio: 0.3,
      },
    });
    expect(candidateEntity.id.length).toBe(36);
  });
  it('should set a candidate entity carpool path', () => {
    const candidateEntity: CandidateEntity = CandidateEntity.create({
      id: '5600ccfb-ab69-4d03-aa30-0fbe84fcedc0',
      role: Role.PASSENGER,
      driverWaypoints: [
        {
          lat: 48.689445,
          lon: 6.17651,
        },
        {
          lat: 48.8566,
          lon: 2.3522,
        },
      ],
      passengerWaypoints: [
        {
          lat: 48.689445,
          lon: 6.17651,
        },
        {
          lat: 48.8566,
          lon: 2.3522,
        },
      ],
      driverDistance: 350145,
      driverDuration: 13548,
      driverSchedule: [
        {
          day: 0,
          time: '07:00',
          margin: 900,
        },
      ],
      passengerSchedule: [
        {
          day: 0,
          time: '07:10',
          margin: 900,
        },
      ],
      spacetimeDetourRatio: {
        maxDistanceDetourRatio: 0.3,
        maxDurationDetourRatio: 0.3,
      },
    }).setCarpoolPath([
      {
        point: new Point({
          lat: 48.689445,
          lon: 6.17651,
        }),
        actors: [
          new Actor({
            role: Role.DRIVER,
            target: Target.START,
          }),
          new Actor({
            role: Role.PASSENGER,
            target: Target.START,
          }),
        ],
      },
      {
        point: new Point({
          lat: 48.8566,
          lon: 2.3522,
        }),
        actors: [
          new Actor({
            role: Role.DRIVER,
            target: Target.FINISH,
          }),
          new Actor({
            role: Role.PASSENGER,
            target: Target.FINISH,
          }),
        ],
      },
    ]);
    expect(candidateEntity.getProps().carpoolSteps).toHaveLength(2);
  });
  it('should create a new candidate entity with spacetime metrics', () => {
    const candidateEntity: CandidateEntity = CandidateEntity.create({
      id: 'cc260669-1c6d-441f-80a5-19cd59afb777',
      role: Role.DRIVER,
      driverWaypoints: [
        {
          lat: 48.678454,
          lon: 6.189745,
        },
        {
          lat: 48.84877,
          lon: 2.398457,
        },
      ],
      passengerWaypoints: [
        {
          lat: 48.689445,
          lon: 6.17651,
        },
        {
          lat: 48.8566,
          lon: 2.3522,
        },
      ],
      driverDistance: 350145,
      driverDuration: 13548,
      driverSchedule: [
        {
          day: 0,
          time: '07:00',
          margin: 900,
        },
      ],
      passengerSchedule: [
        {
          day: 0,
          time: '07:10',
          margin: 900,
        },
      ],
      spacetimeDetourRatio: {
        maxDistanceDetourRatio: 0.3,
        maxDurationDetourRatio: 0.3,
      },
    }).setMetrics(352688, 14587);
    expect(candidateEntity.getProps().distance).toBe(352688);
    expect(candidateEntity.getProps().duration).toBe(14587);
  });
  it('should not validate a candidate entity with exceeding distance detour', () => {
    const candidateEntity: CandidateEntity = CandidateEntity.create({
      id: 'cc260669-1c6d-441f-80a5-19cd59afb777',
      role: Role.DRIVER,
      driverWaypoints: [
        {
          lat: 48.678454,
          lon: 6.189745,
        },
        {
          lat: 48.84877,
          lon: 2.398457,
        },
      ],
      passengerWaypoints: [
        {
          lat: 48.849445,
          lon: 6.68651,
        },
        {
          lat: 47.18746,
          lon: 2.89742,
        },
      ],
      driverDistance: 350145,
      driverDuration: 13548,
      driverSchedule: [
        {
          day: 0,
          time: '07:00',
          margin: 900,
        },
      ],
      passengerSchedule: [
        {
          day: 0,
          time: '07:10',
          margin: 900,
        },
      ],
      spacetimeDetourRatio: {
        maxDistanceDetourRatio: 0.3,
        maxDurationDetourRatio: 0.3,
      },
    }).setMetrics(458690, 13980);
    expect(candidateEntity.isDetourValid()).toBeFalsy();
  });
  it('should not validate a candidate entity with exceeding duration detour', () => {
    const candidateEntity: CandidateEntity = CandidateEntity.create({
      id: 'cc260669-1c6d-441f-80a5-19cd59afb777',
      role: Role.DRIVER,
      driverWaypoints: [
        {
          lat: 48.678454,
          lon: 6.189745,
        },
        {
          lat: 48.84877,
          lon: 2.398457,
        },
      ],
      passengerWaypoints: [
        {
          lat: 48.849445,
          lon: 6.68651,
        },
        {
          lat: 47.18746,
          lon: 2.89742,
        },
      ],
      driverDistance: 350145,
      driverDuration: 13548,
      driverSchedule: [
        {
          day: 0,
          time: '07:00',
          margin: 900,
        },
      ],
      passengerSchedule: [
        {
          day: 0,
          time: '07:10',
          margin: 900,
        },
      ],
      spacetimeDetourRatio: {
        maxDistanceDetourRatio: 0.3,
        maxDurationDetourRatio: 0.3,
      },
    }).setMetrics(352368, 18314);
    expect(candidateEntity.isDetourValid()).toBeFalsy();
  });
});
